#!/usr/bin/env node

var args = require('yargs'),
    M = require('mstring'),
    pkg = require('../package.json'),
    EnvController = require('../lib/controllers/EnvironmentController'),
    AppController = require('../lib/controllers/AppController'),
    prompt = require('prompt'),
    path = require('path'),
    viewHelper = require('../lib/views');


var usage = M(function() {

    /***

     Local Commands:
        new-app                  <app-name>     [-t -m -e -n -u -p]
        view-profile-def                                                                            *
        add-type-to-profile      <type>         [--select --update --insert --aggregate --delete]   *
        remove-type-from-profile <type>                                                             *
        edit-title               <title>                                                            *
        view-title                                                                                  *
        edit-description         <description>                                                      *
        view-description                                                                            *
        edit-version             <version>                                                          *
        view-version                                                                                *


     Remote Commands:
        publish-app                             [-e -n -u -p]                                       *
        get-app-metadata                        [-e -n -u -p]                                       *
        list-apps                               [-e -n -u -p]
        get-profile                             [-e -n -u -p]                                       *
        assign-profile-to-user  <username>      [-e -n -u -p]                                       *

     * Command needs to be run in a specific app directory

     ***/

});

// TODO: make this better!
var argv = args
    .usage(usage)
    .options('t', {alias:'type', describe: 'The project type (custom | angular | app-generator).', default: 'custom'})
    .options('m', {alias:'model', describe: 'The resource type for the app-generator.'})
    .options('e', {alias:'environment', describe: 'The environment to use. Not needed if using default.'})
    .options('n', {alias:'namespace',   describe: 'The namespace. Not needed if environment specified.'})
    .options('u', {alias:'user',    describe: 'The username. Not needed if environment specified.'})
    .options('p', {alias:'password',    describe: "The password. Not needed if environment specified."})
    .options('select', {describe: 'Allow the select action on the type.'})
    .options('insert', {describe: 'Allow the insert action on the type.'})
    .options('delete', {describe: 'Allow the delete action on the type.'})
    .options('update', {describe: 'Allow the update action on the type.'})
    .options('aggregate', {describe: 'Allow the aggregate action on the type.'})
    .options('d', {alias:'debug', type: 'boolean', describe: 'Show debug output', default: false})
    .options('v', {alias:'verbose', type: 'boolean', describe: 'more detailed output', default: false})
    .argv;


//Just want the version
if(argv.version){
    console.log(pkg.version);
    return;
}

//Just want the help
if(argv.help){
    args.showHelp();
    return;
}

function AppTools(options) {

    this.cwd          = process.cwd();   // save this before loadProjectSync changes it
    this.helpers      = require('../lib/helpers')(argv);
    this.output      = require('../lib/views/console')(argv);

    this.project      = this.helpers.loadProjectSync();
    this.createEnv    = this.helpers.createEnv;

    this.controllers  = {
        env: EnvController(this.project, options)
    }
}

AppTools.prototype = {

    // Parse out command
    dispatch: function(argv) {
        var action = (argv._ && argv._.length > 0) ? argv._.shift() : '';
        this.action = action.toLowerCase();
//        this.verbose = argv.v;

        switch(action){
            case '':
                this.output.fatal('No command was found.');
                break;
            case 'new-app':             // done
                this.newApp(argv);
                break;
            case 'publish-app':         // done
                this.publishApp(argv);
                break;
            case 'view-profile-def':
                this.viewProfile(argv);
                break;
            case 'add-type-to-profile':
                this.addTypeToProfile(argv);
                break;
            case 'edit-title':                   // TODO
                this.setTitle(argv);
                break;
            case 'view-title':                   // TODO
                this.getTitle();
                break;
            case 'view-description':             // TODO
                this.getDescription();
                break;
            case 'edit-description':             // TODO
                this.setDescription(argv);
                break;
            case 'view-version':                 // TODO
                this.getVersion();
                break;
            case 'edit-version':                 // TODO
                this.setVersion(argv);
                break;
            case 'get-app-metadata':            // TODO
                this.getAppMetadata();
                break;
            case 'remove-type-from-profile':
                this.removeTypeFromProfile(argv);
                break;
            case 'get-profile':       // done
                this.listProfile(argv);
                break;
            case 'assign-profile-to-user':
                this.assignProfileToUser(argv);
                break;
            case 'list-apps':           // done => need to fix output
                this.listApps(argv);
                break;
            case 'create-app-metadata': // todo
                // TODO
                    // this is for apps not being hosted in cloud
                    // NOTE: need to add custom location (url) where app resides
                break;
            default:
                this.output.fatal('Unknown Command: ' + action + '.');
        }

    },

    // Commands
    newApp: function(argv) {    // Args: -t
                                    // need: -m, -e, -n, -u, -p for app generator

        // Parse inputs
        var name = argv._.shift(),
            type = argv.t || 'custom',
            model = argv.m;

        // Check commands and its required arguments:
        if (!name) this.output.fatal('Missing project name.');

        var nameValid = isNameValid(name);
        if (typeof nameValid === 'string') this.output.fatal(nameValid);

        if (typeof type === 'string') {
            type = type.toLowerCase();
        }

        if (type === 'app-generator' && (!model || (model && typeof model !== 'string'))) {
            this.output.fatal('Please specify a type with the -m flag.');
        }

        var pathToAppFolder = path.join(process.cwd(), 'apps', name);

        this.demandProjectOrExit();

        if (this.helpers.checkIfAppExists(pathToAppFolder)) {
            this.output.fatal('App name already exists! Please choose a different name. (NOTE: app name is not case sensitive)');
        }

        var self = this;

        // Generate skeleton project
        switch (type) {
            case 'custom':

                self.helpers.promptForNewApp()
                    .then(function(result) {

                        self.helpers.createAppDirectory(pathToAppFolder);
                        return self.helpers.createCustomApp(name, result.title, result.description, pathToAppFolder)
                    }, function(err) {
                        self.output.fatal(err);
                    })
                    .then(function(msg) {
                        self.output.info(msg);
                    }, function(msg) {
                        self.output.fatal(msg);
                    });
                break;
            case 'app-generator':

                var credentials = self.helpers.validateCredentials(self.project, argv);
                var data;
                // check if specified type actual exists
                self.helpers.getType(model, credentials)
                    .then(function(response) {
                        data = response;
                        return self.helpers.promptForNewApp();
                    })
                    .then(function(result) {
                        self.helpers.createAppDirectory(pathToAppFolder);
                        return self.helpers.createAppGeneratorApp(model, name, result.title, result.description, pathToAppFolder, data);
                    })
                    .then(function(msg) {

                        self.output.info(msg);

                        // TODO: abstract this out!
                        self.app = self.helpers.loadPackageSync(pathToAppFolder);
                        self.controllers.app = AppController(self.app);

                        var profile = self.controllers.app.addTypeToProfile(model, {select: true});
                        self.app.save();

                        if (!profile.success) {
                            self.output.fatal(profile.message);
                        }

                    }, function(errorMessage) {
                        self.output.fatal(errorMessage);
                    });

                break;
            case 'angular':

                self.helpers.promptForNewApp()
                    .then(function(result) {

                        self.helpers.createAppDirectory(pathToAppFolder);
                        return self.helpers.createAngularSeedApp(name, result.title, result.description, pathToAppFolder);
                    })
                    .then(function(msg) {

                        self.output.info(msg);
                    }, function(msg) {

                        self.output.fatal(msg);
                    });
                break;
            default:
                // in case user spells it wrong...
                self.output.fatal('Unknown project type. Please specify either custom, angular, or app-generator.');
        }

//        });

        /** HELPER FUNCTION **/
        function isNameValid(name) {
            var length = name.length,
                matches = name.match(/[a-zA-Z0-9_-]/g);

            if (length < 3) {

                return 'App name has to have at least 3 characters.'
            } else if (!matches || length !== matches.length) {

                return 'App name can only contain [a-zA-Z0-9_-]';
            } else {

                return true;
            }

        }

    },

    // ASSUMPTION: you have to be in the app directory to publish!
    publishApp: function(argv) {    // Possible Args: -e -n -u -p --dir (optional -> default to /www)
        // ENDPOINT: PUT or POST to /apps

        var view = viewHelper.View(viewHelper.createSpec(argv));

        // Parse inputs
        var publishDirectory = 'www';

        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);

        // Validate
        this.helpers.checkFile(pathToAppFolder, publishDirectory, 'index.html');
        this.helpers.checkFile(pathToAppFolder, publishDirectory, 'package.json');
        this.helpers.checkFile(pathToAppFolder, publishDirectory, 'LICENSE');

        // Create app instance
        // NOTE: for this command only we will use the package.json from the publish directory
        this.app = this.helpers.loadPackageSync(path.join(pathToAppFolder, publishDirectory));
        this.controllers.app = AppController(this.app);

        // Make sure a project exists so can have an env
        this.demandProjectOrExit();

        // Validate credentials
        var credentials = this.helpers.validateCredentials(this.project, argv);

        // Make sure all the required attributes are present before publishing
        var validApp = this.controllers.app.validateApp();
        if (!validApp.valid) {
            this.output.fatal(validApp.msg);

            // TODO: prompt for missing required fields?
        }

        // Save overwrites to package.json
        this.app.save();

        // Create Zip
        var self = this;
        this.helpers.createAppZip(this.app.getName(), pathToAppFolder, publishDirectory, new Date().getTime())
            .then(function(obj) {

                var zipName = obj.zipName,
                    size    = obj.size;

                self.output.verbose(obj.msg);

                // Publish App
                return self.helpers.sendApp(zipName, size, pathToAppFolder, credentials, self.app);

            })
            .then(function(response) {

                self.output.info(response.message);

                // Assign Profile to publisher
                if (response.type === 'POST') {

                    self.helpers.assignProfile(self.app.getProfile().name, credentials.username, credentials)
                        .then(function(msg) {

                            self.output.info(msg);
                        }, function(msg) {

                            self.output.fatal(msg);
                        });
                }

                // Get app meta data to make sure it published
                self.helpers.getAppMetadata(credentials, self.app.getName())
                    .then(function(response) {

                        view.render(response);
                    }, function(err) {

                        self.output.fatal(err);
                    });

            }, function(msg) {

                self.output.fatal(msg.error);
            })
            .fail(function(msg) {

                self.output.fatal(msg);
            });

    },

    getAppMetadata: function() {

        var view = viewHelper.View(viewHelper.createSpec(argv));

        var self = this;

        // Make sure a project exists so can have an env
        this.demandProjectOrExit();

        // Get credentials
        var credentials = this.helpers.validateCredentials(this.project, argv);

        // Get app attributes from package.json
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);
        this.app = this.helpers.loadPackageSync(path.join(pathToAppFolder));

        this.helpers.getAppMetadata(credentials, this.app.getName())
            .then(function(response) {

                view.render(response);
//                self.output.info(response);
            }, function(err) {

                self.output.fatal(err);
            });
    },

    // NOT IN USE
    removeAppRemote: function(argv) {

        // ENDPOINT: DELETE /apps/<app name>

        // TODO
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);

        // Create app instance
        this.app = this.helpers.loadPackageSync(pathToAppFolder);
        this.controllers.app = AppController(this.app);
        var name = this.app.getName();

        // Make sure a project exists so can have an env
        this.demandProjectOrExit();

        // Validate credentials
        var credentials = this.helpers.validateCredentials(this.project, argv);

        // Make sure all the required attributes are present before publishing
        var validApp = this.controllers.app.validateApp();
        if (!validApp.valid) {
            this.output.fatal(validApp.msg);
//            this.helpers.fatal('Missing required fields in package.json. Note: both name and version are required.');

            // TODO: prompt for missing required fields?
        }

        var self = this;

        this.helpers.removeAppRemote(credentials, name)
            .then(function() {

                console.log('Successfully removed ' + name + ' from the server.');

            }, function(err) {

                self.helpers.fatal('Failed to remove ' + name + ' from the server.');

            });
    },

    packageCurrier: function() {

        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd),
            app = this.helpers.loadPackageSync(pathToAppFolder);

        return function(setOrGet, attribute, value) {

            if (setOrGet === 'set') {

                app[attribute](value);
                app.save();
                return value;
            } else if (setOrGet === 'get') {

                return app[attribute]();
            }
        };
    },

    setTitle: function(argv) {

        // Parse options/arguments
        var title = argv._.shift();

        // Error handling
        if (argv._.length > 0) this.output.fatal('Only one argument allowed.' +
            ' Specified title must also be wrapped in quotes!');

        this.packageCurrier()('set', 'setTitle', title);

        this.output.info('Title set to: ' + title);
    },

    getTitle: function() {

        this.output.info(this.packageCurrier()('get', 'getTitle'));
    },

    setDescription: function(argv) {

        var description = argv._.shift();

        if (argv._.length > 1) this.output.fatal('Only one argument allowed.' +
            ' Specified description must also be wrapped in quotes!');

        this.packageCurrier()('set', 'setDescription', description);
        this.output.info('Description set to: ' + description);
    },

    getDescription: function() {

        this.output.info(this.packageCurrier()('get', 'getDescription'));
    },

    setVersion: function(argv) {

        var version = argv._.shift();

        if (!version) this.output.fatal('Version was not specified.');

        this.packageCurrier()('set', 'setVersion', version);
        this.output.info('Version set to: ' + version);
    },

    getVersion: function() {

        this.output.info(this.packageCurrier()('get', 'getVersion'));
    },

    // Note this is local
    viewProfile: function(argv) {
        var view = viewHelper.View(viewHelper.createSpec(argv));

        // Create app instance
        // TODO: abstract this out into a fn!!
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);
        this.app = this.helpers.loadPackageSync(pathToAppFolder);
        this.controllers.app = AppController(this.app);

        var profile = this.controllers.app.viewProfile();

        if (profile.success) {

            if (argv.color) {
                view.render(profile.profile);
            } else {
                this.output.info(JSON.stringify(profile.profile, null, 4));
            }

//            this.output.info(JSON.stringify(profile.profile, null, 4));
//            this.output.info(view.render(profile.profile));
        } else {
            this.output.fatal(profile.message);
        }

    },

    addTypeToProfile: function(argv) {

        var view = viewHelper.View(viewHelper.createSpec(argv));

        // Parse arguments
        var typeName = argv._.shift(),
            actions = {
                select: argv.select,
                insert: argv.insert,
                delete: argv.delete,
                aggregate: argv.aggregate,
                update: argv.update
            };

        // Validate arguments/options
        if (!typeName) {

            this.output.fatal('Please provide a type.');
        }
        if (!actions.select &&
            !actions.insert &&
            !actions.delete &&
            !actions.aggregate &&
            !actions.update) {

            this.output.fatal('Please provide at least one action (select, insert, delete, aggregate, update).');
        }

        // Create app instance
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);
        this.app = this.helpers.loadPackageSync(pathToAppFolder);
        this.controllers.app = AppController(this.app);

        var profile = this.controllers.app.addTypeToProfile(typeName, actions);
        this.app.save();

        if (profile.success) {

            // TODO: ideally don't have to do the check... view.render should just handle it...
            if (argv.color) {
                view.render(profile.profile);
            } else {
                this.output.info(JSON.stringify(profile.profile, null, 4));
            }

        } else {
            this.output.fatal(profile.message);
        }

    },

    removeTypeFromProfile: function(argv) {

        var view = viewHelper.View(viewHelper.createSpec(argv));

        var typeName = argv._.shift();

        // Create app instance
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);
        this.app = this.helpers.loadPackageSync(pathToAppFolder);
        this.controllers.app = AppController(this.app);

        var profile = this.controllers.app.removeTypeFromProfile(typeName);

        this.app.save();

        if (profile.success) {

            if (argv.color) {
                view.render(profile.profile);
            } else {
                this.output.info(JSON.stringify(profile.profile, null, 4));
            }
        } else {
            this.output.fatal(profile.message);
        }
    },

    assignProfileToUser: function(argv) {

        var username = argv._.shift();

        // Make sure a project exists so can have an env
        this.demandProjectOrExit();

        // Create app instance
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);
        this.app = this.helpers.loadPackageSync(pathToAppFolder);

        // Validate credentials
        var credentials = this.helpers.validateCredentials(this.project, argv);

        var self = this;
        this.helpers.assignProfile(this.app.getProfile().name, username, credentials)
            .then(function(msg) {
                self.output.info(msg);
            });

        // TODO: return user record?
    },

    // NOT IN USE
    addProfile: function(argv) {

        // ENDPOINT: POST /apps/<app name>/profiles/<profile name (KFC.user)>

        this.profileCurrier('add')(argv);

    },

    // NOT IN USE
    removeProfile: function(argv) {

        // ENDPOINT: DELETE /apps/<app name>/<profile name>

        this.profileCurrier('remove')(argv);
    },

    listProfile: function(argv) {

        // ENDPOINT: GET /apps/<app name>

        this.profileCurrier('list')(argv);
    },

    profileCurrier: function(type) {

        var self = this,
            successMessage = '',
            failMessage = '',
            fn;

        switch(type) {
            case 'add':
                fn = this.helpers.addProfile;
                successMessage = 'Successfully added profile';
                failMessage = 'Failed to add Profile.';
                break;
            case 'remove':
                fn = this.helpers.removeProfile;
                successMessage = 'Successfully removed profile';
                failMessage = 'Failed to remove Profile.';
                break;
            case 'list':
                fn = this.helpers.listProfile;
                failMessage = 'Failed to get profile.';
                break;
        }

        return function(argv) {

            var profileName;

            if (type === 'remove' || type === 'add') {
                profileName = argv._.shift();
                if (!profileName) {
                    self.output.fatal('Missing profile name. Please specify one.');
                }
            }

            self.demandProjectOrExit();

            // Validate credentials
            var credentials = self.helpers.validateCredentials(self.project, argv);

            var pathToAppFolder = self.helpers.getPathToAppFolder(self.cwd);
            self.app = self.helpers.loadPackageSync(pathToAppFolder);

            fn(self.app.getName(), credentials, profileName)
                .then(function(profile) {

                    if (profile) {
                        self.output.info(profile);

                        // TODO: show dml of profile!
                    } else {
                        self.output.info(successMessage);
                    }

                }, function(err) {
                    self.output.fatal(err);
                });
        }
    },

    listApps: function(argv) {

        // ENDPOINT: Get /apps

        var credentials = this.helpers.validateCredentials(this.project, argv),
            self = this,
            view = viewHelper.View(viewHelper.createSpec(argv));

        this.helpers.listApps(credentials)
            .then(function(apps) {

                view.render(apps);
//                self.output.info(apps);
//                self.output.info(view.render(apps));
            }, function() {
                self.output.fatal('Failed to get apps. Please try again.');
            });
    },

    // NOTE: these are repeats... abstract this out for all the commands?
    demandProjectOrExit: function(){
        if(!this.project){
            this.output.fatal('Unknown project. Please init a new project in this directory or move to an active project.');
        }
    }
};

// MAIN
new AppTools(argv).dispatch(argv);