#!/usr/bin/env node

var args                 = require('yargs')
  , M                    = require('mstring')
  , fs                   = require('fs')
  , path                 = require('path')
  , pkg                  = require('../package.json')
  , helpers              = require('../lib/helpers.js')
  , ProjectController    = require('../lib/controllers/ProjectController')
  , EnvController        = require('../lib/controllers/EnvironmentController')
  , RestController       = require('../lib/controllers/RestController')
  , viewHelper           = require('../lib/views');

    var usage = M(function(){
    /***
         $0 [options] <command> [args]

         //- Project Commands
         pwd         |
         init        |
         set-default |          <env-name>
         get-default |

         //- Env Commands
         new         | create   <env-name>
         ls          | list
         view        | print    <env-name>
         edit        | modify   [env-overloads]
         mv          | rename   <from> <to>  [env-overloads]
         cp          | clone    <from> <to>  [env-overloads]
         rm          | delete   <env-name>

         //- Env Details
         user        | whoami   [-e <env-name>]
         url         | uri      [-e <env-name>]
         namespace   |          [-e <env-name>]
         whereami    |          [-e <env-name>]

         //- Http Options
         ping        |          [-e <env-name>] [env-overloads]
         login       | authn    [-e <env-name>] [env-overloads]
         get         |          [-e <env-name>] [env-overloads] <path> [--out]
         count       |          [-e <env-name>] [env-overloads] <path> [--out]
         put         |          [-e <env-name>] [env-overloads] <path> [--out]
         post        |          [-e <env-name>] [env-overloads] <path> [--out]
         patch       |          [-e <env-name>] [env-overloads] <path> [--out]
         delete      |          [-e <env-name>] [env-overloads] <path> [--out]
         head        |          [-e <env-name>] [env-overloads] <path> [--out]
         options     |          [-e <env-name>] [env-overloads] <path> [--out]

         --view      { status, headers, location, id, entity, full, debug }
         --color
         --flat
         --status    include the status code

         ***/});

    var argv = args
        .usage(usage)
        .options('V', {alias:'version',     describe: 'Print version number of the tool'})
        .options('e', {alias:'environment', describe: 'The environment to use'})
        .options('s', {alias:'uri',         describe: 'The service provider uri'})
        .options('u', {alias:'username',    describe: 'The remote username'})
        .alias('u', 'user')
        .options('p', {alias:'password',    describe: "The remote user's password"})
        .options('n', {alias:'namespace',   describe: 'The active namespace'})
        .options('a', {alias:'agent',       describe: 'The client user agent'})
        .options('D', {alias:'data',        describe: 'A data string to supply'})
        .options('F', {alias:'file',        describe: 'The file that contains the json'})
        .options('o', {alias:'out',         describe: 'Describe the output'})
        .options('view'   , {describe: 'Describe the view'})
        .options('color'  , {describe: 'present the view in console colors', type:'boolean', default: false})
        .options('flat'   , {describe: 'provide flat json structure', type:'boolean', default: false})
        .options('status' , {describe: 'include the status', type:'boolean', default: false})
        .options('d', {alias:'debug',       type: 'boolean', describe: 'Show debug output'   ,       default: false})
        .options('f', {alias:'force',       type: 'boolean', describe: 'Force the command'   ,       default: false})
        .options('v', {alias:'verbose',     type: 'boolean', describe: 'more detailed output',       default: false})
        .argv
        ;

//Just want the version
if(argv.version){
    console.log(pkg.version);
    return;
}

//Just want the help
if(argv.help){
    args.showHelp()
    return;
}

function Cli(options){
    this.helpers = helpers(options)

    //Map functions from helpers
    this.project = this.helpers.loadProjectSync();
    this.createEnvPatch = this.helpers.createEnvPatch;

    //Command Controllers
    this.controllers = {
        project: ProjectController(this.project, options),
        env    : EnvController(this.project, options),
        rest   : RestController(this.project, options)
    }
}

Cli.prototype = {

    dispatch: function(argv){

        //Consume the action
        var action = (argv._ && argv._.length > 0) ? argv._.shift() : '';
        this.action = action.toLowerCase()
        switch(this.action){
            case 'info':
                if(argv.debug) console.log('info');
                this.info(argv);
                break;
            case 'pwd':
                if(argv.debug) console.log('pwd');
                this.pwd(argv);
                break;
            case 'init':
                if(argv.debug) console.log('init')
                this.init(argv);
                break;
            case 'set-default':
                if(argv.debug) console.log('set-default')
                this.setDefault(argv);
                break;
            case 'get-default':
                if(argv.debug) console.log('get-default')
                this.getDefault(argv);
                break;
            case 'new':
            case 'create':
                if(argv.debug) console.log('create')
                this.create(argv);
                break;
            case '':
            case 'ls':
            case 'll':
            case 'dir':
            case 'list':
                if(argv.debug) console.log('list');
                this.list(argv)
                break;
            case 'view':
            case 'print':
                if(argv.debug) console.log('view')
                this.view(argv);
                break;
            case 'edit':
            case 'modify':
                if(argv.debug) console.log('edit')
                this.edit(argv);
                break;
            case 'rm':
            case 'remove':
                if(argv.debug) console.log('remove')
                this.remove(argv);
                break;
            case 'cp':
            case 'copy':
                if(argv.debug) console.log('copy')
                this.copy(argv);
                break;
            case 'mv':
            case 'move':
            case 'rename':
                if(argv.debug) console.log('rename')
                this.rename(argv);
                break;
            case 'user':
            case 'whoami':
                if(argv.debug) console.log('user')
                this.view(argv, 'username');
                break;
            case 'uri':
            case 'url':
                if(argv.debug) console.log('uri');
                this.view(argv, 'uri');
                break;
            case 'namespace':
                if(argv.debug) console.log('namespace');
                this.view(argv, 'namespace');
                break;
            case 'whereami':
                this.view(argv, 'location');
                break;
            case 'ping':
                this.ping(argv);
                break;
            case 'login':
            case 'auth':
                this.auth(argv);
                break;
            case 'get':
            case 'count':
            case 'put':
            case 'post':
            case 'patch':
            case 'delete':
            case 'head':
            case 'options':
                if(argv.debug) console.log('rest call')
                this.request(argv)
                break;
            default:
                console.log('Unknown Command: ' + action);
                break;
        }

    },

    info: function(){

        var logo;
        try{
            logo = fs.readFileSync(path.join(__dirname, '..', 'lib', 'assets', 'logo.txt'), 'utf-8');
        } catch (e){}

        console.log(pkg.name + ' v' + pkg.version);
        if(logo) console.log(logo);
        console.log(pkg.description);
        console.log('Licensed under ' + pkg.license);
        console.log('\u00A9 2015 HotSchedules Inc.');
    },

    pwd: function(){
        this.controllers.project.pwd();
    },

    init: function(){
        this.controllers.project.init();
    },

    getDefault: function(){
        this.demandProjectOrExit();
        this.controllers.project.getDefault();
    },

    setDefault: function(argv){

        this.demandProjectOrExit();

        if(argv._.length > 1){
            console.log("Usage: > rbc-cli set-default <environment-name>")
            return;
        }

        var name = argv._.shift();

        if(!this.project){
            console.error('Unknown project. Please init or change to a working project.');
            return
        }

        this.controllers.project.setDefault(name);

    },

    create: function(argv){

        this.demandProjectOrExit();

        if(argv._.length > 1){
            console.log("Usage: > rbc-cli create <environment-name> [env-options]")
            return;
        }

        var name = argv._.shift();
        var spec = this.createEnvPatch(argv);

        this.controllers.env.create(name, spec);
    },

    /**
     * List the environments
     * @param argv
     */
    list: function(argv){
        this.demandProjectOrExit();
        this.controllers.env.list();
    },

    view: function(argv, view){

        this.demandProjectOrExit();

        var name;

        if(argv._ && argv._.length){
            name = argv._.shift();
        } else if(argv.environment){
            name = argv.environment;
        } else {
            name = null;
        }

        this.controllers.env.view(name, view);
    },

    rename: function(argv){

        this.demandProjectOrExit();

        var from    = argv._.shift();
        var to      = argv._.shift();
        var patch   = this.createEnvPatch(argv);

        this.controllers.env.rename(from, to, patch);

    },

    copy: function(argv){

        this.demandProjectOrExit();

        var from    = argv._.shift();
        var to      = argv._.shift();
        var patch   = this.createEnvPatch(argv);

        this.controllers.env.copy(from, to, patch);
    },

    edit: function(argv){

        this.demandProjectOrExit();

        var name    = argv._.shift();
        var patch   = this.createEnvPatch(argv);

        this.controllers.env.update(name, patch);
    },

    remove: function(argv){

        this.demandProjectOrExit();

        var name = argv._.shift();

        this.controllers.env.remove(name);
    },

    ping: function(argv){
        this.demandProjectOrExit();

        var env       = this.helpers.createEnv(this.project, argv);
        var action    = 'GET';
        var resource  = '/';
        var json      = null;
        var output    = 'ping';
        var saveAs    = null;
        var view      = viewHelper.createSpec(argv);
        view.view     = view.view || output;

        this.controllers.rest.request(env, action, resource, json, view);
    },

    auth: function(argv){

        this.demandProjectOrExit();

        var env         = this.helpers.createEnv(this.project, argv);
        var action      = 'GET';
        var resource    = '/me';
        var json        = null;
        var output      = 'login';
        var saveAs      = null;

        var view        = viewHelper.createSpec(argv);
        view.view       = view.view || output;
        view.drop       = "password";

        this.controllers.rest.request(env, action, resource, json, view);

    },

    request: function(argv){

        this.demandProjectOrExit();

        var env    = this.helpers.createEnv(this.project, argv);
        var action = this.action;
        var resource  = argv._.shift();
        var input  = this._toJson(argv);
        var view   = viewHelper.createSpec(argv);


        //- Controller
       this.controllers.rest.request(env, action, resource, input, view);

    },

    _toJson: function(argv){
        var json = null;
        if((argv.data)){
            try{
                json = JSON.parse(options.data);
            } catch (err) {
                console.log('Invalid JSON');
                console.log(err.message);
            }
        } else if(argv.file){
            try{
                json = JSON.parse(fs.readFileSync(argv.file, 'utf8'));
            } catch (err) {
                console.log('Invalid JSON');
                console.log(err.message);
            }
        }
        return json;

    },

    demandProjectOrExit: function(){
        if(!this.project){
            this.fatal('Unknown project. Please init a new project in this directory or move to an active project.');
        }
    },

    fatal: function (msg, code){
        code = code || 1;
        console.log('Fatal Error' + ' ' + msg);
        process.exit(code);
    }

}

new Cli(argv).dispatch(argv);