module.exports = function(project, options){

    var _  = require('../utils.js')
      , path    = require('path')
      , fs      = require('fs-extra')
      , Resource= require('../types').Resource
      , Embedded= require('../types').Embedded
      , Property= require('../types').Property
    ;

    /**
     * Create the model
     */
    function create(type_name, spec){

        //map
        var name        = type_name;
        var storeAs     = spec.storage_name || name;
        var type;

        //Belongs in a processor
        if(spec.embedded){
            type = new Embedded(type_name);
        } else {
            type = new Resource(type_name);
            type.storedAs(storeAs);
            if(spec.immutable) type.asImmutable();
            if(spec.extensible) type.asExtensible();
            if(spec.encapsulated) type.asEncapsulated();
        }

        save(type,  function done(err){
            if(err){
                console.log('failed to create the type %s', type_name);
            } else {
                console.log('created the type %s', type_name);
            }
        });
    }

    function copy(from, to){
        return rename(from, to, false);
    }

    function move(from, to){
        return rename(from, to, true);
    }

    function rename(from, to, destroy){

        //
        if(from == null){
            return console.log('Invalid Argument: failed to specify existing type name.')
        }

        if(to == null){
            return console.log('Invalid Argument: failed to specify new type name.')
        }

        //Belongs in the processor
        var type = load(asFilePath(from));
        if(!type){
            return console.log('failed to load local type %s', from);
        }

        type.name = to;

        if(type.storage_name === from){
            type.storage_name = to;
        }

        save(type,  function done(err){
            if(err){
                console.log('failed to rename the type %s', from);
            } else {
                if(destroy) remove(from);
                console.log('created the type %s from %s', to, from);
            }
        });
    }

    /**
     * Show a local model
     *
     * -color    : print in color
     * -minimal  : only show the required fields
     * -flat     : do not indent
     *
     */
    function view(type_name, view){
        try{
            var model = fs.readJsonFileSync(asFilePath(type_name));
            if(model) {
                view.render(model);
            } else {
                console.log('unable to render the existing model')
            }
        } catch (err){
            console.log('failed to find the local type %s', type_name)
        }
    }

    /**
     * Get the model from the server
     */
    function edit(type_name, patch){

        //Push Down to the type models
        var type = load(asFilePath(type_name));
        if(!type){
            return console.log('failed to load local type %s', type_name);
        }

        //-Converting to embedded
        if(patch.embedded){
            type.embedded = true;
            delete type.storage_name;
            delete type.immutable;
            delete type.encapsulated;
            delete type.extensible;
        } else {
            if(patch.storage_name)  type.storage_name = patch.storage_name;
            if(patch.immutable)     type.storage_name = patch.immutable;
            if(patch.encapsulated)  type.storage_name = patch.encapsulated;
            if(patch.extensible)    type.storage_name = patch.extensible;
        }

        save(type,  function done(err){
            if(err){
                console.log('failed to modify the type %s', type_name);
            } else {
                console.log('modified the type %s', type_name);
            }
        });
    }

    function asFilePath(name){
        return path.join(process.cwd(), 'models', (name + '.json'));
    }

    function load(type_name){
        try{
            var raw = fs.readJsonSync(type_name);
            var type = (raw.embedded) ? new Embedded() : new Resource();
            type.merge(raw);
            return  type;
        } catch (err) {
            return  undefined;
        }
    }

    function saved(resource, filepath){
        return function _onSaved(err){
            if(err){
                if(options.verbose) console.log('Unable to write the project file');
            } else {
                console.log(filePath);
            }
        }
    }

    function save(resource, response){

        if(!resource){
            return;
        }

        fs.outputFile(asFilePath(resource.name), _.toDelimitedJson(resource),
                      response || saved(resource, filepath));
    }

    function remove(type_name){
        fs.remove(asFilePath(type_name), function(err){
            if (err) {
                return console.error("failed to delete the type " + type_name);
            } else {
                return console.log('type ' + type_name + ' locally removed');
            }
        });
    }

    function getProps(type_name){
        var type = load(asFilePath(type_name));

        if(!type){
            return console.log('failed to load local type %s', type_name);
        }
        _.forEach(type.properties, function(value, key, lisrt){
            if(!_.startsWith(key, 'sys_')){
                console.log(key);
            }
        })
    }

    function genInstance(type_name){
        var type = load(asFilePath(type_name));
        if(!type){
            return console.log('failed to load local type %s', type_name);
        }

        //get the property names
        var propertyNames = _.keys(type.properties);

        if(!propertyNames){
            return console.log('no property names are available');
        }

        propertyNames.sort();

        var instance = {};

        for(var i = 0; i < propertyNames.length; i++){
            var key  = propertyNames[i];
            var prop = type.properties[key];

            //For all non rbc attributes
            if(!_.startsWith(key, 'sys_')){

                var pType = prop.type.toLowerCase();
                var pValue=null;
                switch(pType){
                    case 'boolean':
                        pValue = false;
                        break;
                    case 'integer':
                        pValue = 0;
                        break;
                    case 'real':
                        pValue = 0.00;
                        break;
                    case 'string':
                        pValue = "";
                        break;
                    case 'datetime':
                        pValue = new Date().toISOString();
                        break;
                    case 'currency':
                        pValue =  {
                            code    : 'USD',
                            scale   : 2,
                            value   : 0
                        };
                        break;
                    case 'geojson':
                        pValue = [0.00, 0.00];
                        break;
                    default:
                        pValue = (pType.multi) ? [] : {};
                        break;
                }
                instance[key] = (prop.multi) ? [pValue]: pValue;
            }
        }
        console.log(_.toDelimitedJson(instance));
    }

    function putProp(type_name, properties, spec){

        var type = load(asFilePath(type_name));
        if(!type){
            return console.log('failed to load local type %s', type_name);
        }


        _.forEach(properties, function(prop_name){

            var prop = new Property(prop_name)

            //Select the type
            if(spec.type){
                prop.ofType(spec.type)
            } else if (spec.Boolean){
                prop.asBoolean();
            } else if (spec.Integer){
                prop.asInteger();
            } else if (spec.Real){
                prop.asReal();
            } else if (spec.String){
                prop.asString();
            } else if (spec.Link){
                prop.asLink();
            } else if (spec.DateTime){
                prop.asDateTime();
            } else if (spec.Currency){
                prop.asCurrency();
            } else if (spec.GeoJSON){
                prop.asGeoJSON();
            } else {
                prop.asObject();
            }

            //Property Options
            if(spec.required)   prop.asRequired(true);
            if(spec.unique)     prop.asUnique(true);
            if(spec.multi)      prop.asList(true);
            if(spec['default']) prop.withDefault(spec['default'])

            type.bindProperty(prop);
        });

        save(type, function done(err){
            if(err){
                console.log('unable to add properties to %s', type_name);
            } else {
                console.log('added properties %s to type %s', properties, type_name);
            }
        });
    }

    function mvProp(type_name, from_name, to_name){

        var type = load(asFilePath(type_name));
        if(!type){
            return console.log('failed to load local type %s', type_name);
        }

        type.properties = type.properties || {};

        if(type.properties[from_name]){
            type.properties[to_name] = type.properties[from_name];
            delete type.properties[from_name];
        }

        save(type, function done(err){
            if(err){
                console.log('failed to rename the property %s on type %s', from_name, type_name);
            } else {
                console.log('renamed the property %s to %s', from_name, to_name);
            }
        });
    }

    function rmProp(type_name, property){

        var type = load(asFilePath(type_name));
        if(!type){
            return console.log('failed to load local type %s', type_name);
        }

        delete type.properties[property];
        save(type,  function done(err){
            if(err){
                console.log('unable to remove the property %s from ', type_name);
            } else {
                console.log('added %s to type %s', properties, type_name);
            }
        });
    }

    return {
        create     : create,
        rename     : move,
        copy       : copy,
        view       : view,
        edit       : edit,
        remove     : remove,
        getProps   : getProps,
        putProp    : putProp,
        mvProp     : mvProp,
        rmProp     : rmProp,
        genInstance: genInstance
    }
};

