module.exports = (function(){
    'use strict'

    var fs          = require('fs-extra');
    var path        = require('path');
    var _           = require('./utils');
    var Environment = require('./environment');


    var pkg         = require('../package.json');

    var PROJECT_FILE    = "rbc-project.json",

        DEFAULT_PROJECT = {
            version      : pkg.version
            , environments : {}
        }

    var noop =  function(){};

    /* - SERIALIZED FORM
     name, description, environments, and options can all be set by the user
     {
         "name"          : "rbc-cli",
         "description"   : "My Project",
         "version"       : "0.1.0",
         "default-env"   : "env"
         "environments"  : {
            env: {
                uri      : <http://host:port/>
                user     : username
                password : secret
                namespace: system
            }
         }
     }
     */

    /**
     * Project - represents serialized state
     * @param attributes
     * @param options
     * @constructor
     */
    function Project(attributes, options){
        options         = options || {}
        this.path       = options.home || process.cwd();
        if(attributes) this.exists = true;
        this.attributes =  _.extend(DEFAULT_PROJECT, attributes);
    }

    Project.prototype = {

        init: function(){
            fs.mkdirpSync(path.join(this.path, 'models'));
            // ADDED - TL
            fs.mkdirpSync(path.join(this.path, 'apps'));
        }

        , getBaseDir : function(){
            return this.path
        }

        , getProjectFile : function(){
            return path.join(this.path, PROJECT_FILE);
        }

        , hasDefaultEnv: function(name){
            return this.attributes['default'] != null;
        }

        , getDefaultEnvironment: function(name){
            return this.attributes['default'];
        }

        , setDefaultEnvironment: function(name){
            this.attributes['default'] = name;
            return this;
        }

        , createEnvironment: function(name, attr){
            name    = name || 'env' + Date.now();
            var env = new Environment(name, attr)
            this.setEnvironment(env);
            return env;
        }

        , getAttributes: function(){
            return this.attributes;
        }

        , getEnvironments: function(){
            return this.attributes.environments;
        }

        /**
         * Has an Connection
         * @param key
         * @returns {boolean}
         */
        , hasEnvironment: function(key){
            return this.getEnvironments()[key] != null;
        }

        /**
         * Get an Connection
         * @param key
         * @returns {boolean}
         */
        , renameEnvironment: function(theOld, theNew){
            if(this.hasEnvironment(theOld)){
                var basis = this.getEnvironments()[theOld];
                this.getEnvironments()[theNew] = _.extend({}, basis)
                delete this.attributes.environments[theOld];
            }
            return this;
        }

        /**
         * Get an Connection
         * @param key
         * @returns {boolean}
         */
        , copyEnvironment: function(basis, clone, patch){
            if(this.hasEnvironment(basis)){
                var basis = this.getEnvironments()[basis];
                this.getEnvironments()[clone] = _.extend({}, basis, patch)
            }
            return this;
        }

        /**
         * Get an Connection
         * @param key
         * @returns {boolean}
         */
        , getEnvironment: function(key){
            return this.hasEnvironment(key) ? new Environment(key, this.getEnvironments()[key])
                                            : null;
        }

        /**
         * Replace an Connection
         *
         * @param key
         * @param env
         */
        , setEnvironment: function(key, env){

            var storeAs, environment;

            if(_.isObject(key)) {
                env = key;
                key = env.name;
            }

            if(!key){
                throw new Error('Illegal Argument: Unnamed Connection');
            }

            this.getEnvironments()[key] = env.getAttributes();
            env.project = this;

            return this;
        }

        /**
         * Partially update the environment with a patch
         * @param key
         * @param patch
         * @returns {Project}
         */
        , patchEnvironment: function(key, patch){
            if(this.hasEnvironment(key)){
                var current = this.attributes.environments[key];
                current = _.assign(current, patch);
                this.attributes.environments[key] = current;
            }
            return this;
        }

        , removeEnvironment: function(key){
            delete this.attributes.environments[key];
        }

        /**
         * Save the project file
         * function(err)
         */
        , save       : function(callback){
            callback = callback || noop;
            fs.outputJson(this.getProjectFile(), this.getAttributes(), callback);
        }

        /**
         * Remove the project file
         * function(err)
         */
        , destroy    : function(callback){
            callback = callback || noop;
            fs.remove(this.getProjectFile(), callback);
        }

        , toString   : function(){
            return "Project@" + this.getProjectFile();
        }

        , toJSON     : function(){
            return _.toDelimitedJson(this);
        }
    }

    return Project

})();