module.exports = (function(){

    //Imports
    var _     = require("lodash");
    var JSON2 = require('JSON2');

    //Primitives
    var _BOOLEAN    = "Boolean",
        _INTEGER    = "Integer",
        _NUMERIC    = "Real",
        _STRING     = "String",
        _DATETIME   = "DateTime",
        _OBJECT     = "Object";

    var _PRIMITIVE_SCALAR = [
        _BOOLEAN, _INTEGER, _NUMERIC, _DATETIME, _STRING, _OBJECT
    ]

    function isPrimitiveScalar(type){
        return (/(_BOOLEAN|_INTEGER|_NUMERIC|_DATETIME|_STRING|_OBJECT)/i.test(type));
    }

    //Scalars
    var _CURRENCY    = "Currency",
        _GEO         = "GeoJSON",
        _LINK        = "Link"
        ;

    function isExtendedScalar(type){
        return (/(_CURRENCY| _DECIMAL)/i.test(type));
    }

    function isStringScalar(type){
        return (/_STRING/i.test(type));
    }

    function toJSON(delimiter) {
        return (delimiter) ? JSON2.stringify(JSON2.decycle(this), null, delimiter)
            : JSON2.stringify(JSON2.decycle(this))
    }

    /**
     * *********
     * FUNCTIONS
     * *********
     */

    function multi(){
        this.multi = true;
        return this;
    }

    var PropertySpec = function(name){
        this.name = name;
        this.type = _OBJECT;
    }

    PropertySpec.prototype = {

        ofType: function(type){
            this.type = type;
            return this;
        },

        asBoolean: function(){
            this.type = _BOOLEAN
            return this;
        },

        asInteger: function(){
            this.type = _INTEGER
            return this;
        },

        asReal: function(){
            this.type = _NUMERIC
            return this;
        },

        asString: function(){
            this.type = _STRING;
            return this;
        },

        asDateTime: function(){
            this.type = _DATETIME;
            return this;
        },

        asCurrency: function(){
            this.type  = _CURRENCY;
            return this;
        },

        asGeoJSON: function(){
            this.type = _GEO;
            return this;
        },

        asLink: function(){
            this.type = _OBJECT;
            return this;
        },

        asObject: function(){
            this.type = _OBJECT;
            return this;
        },

        asRequired: function(){
            this.required = true;
            return this;
        },

        asList: function(){
            this.multi = true;
            return this;
        },

        asUnique: function(option){
            this.unique = true;
            return this;
        },

        withDefault: function(val){
            this["default"] = val;
            return this;
        },

        toJSON: function(delimiter){
            return toJSON.apply(this, arguments)
        },

        toString: function(){
            return "property:" + this.type  + ":" + this.name;
        }

    }

    /**
     * String Property
     * @param name
     * @constructor
     */
    var BooleanProperty = function(name){
        this.name = name;
        this.type = _BOOLEAN;
    }

    _.extend(BooleanProperty.prototype, PropertySpec.prototype, {

        defaultsToFalse: function(){
            this["default"] = true;
        },

        defaultsToTrue : function(){
            this["default"] = true;
        }

    });

    /**
     * Integer Property
     * @param name
     * @constructor
     */
    var IntegerProperty = function(name){
        this.name = name;
        this.type = _INTEGER;
    }

    _.extend(IntegerProperty.prototype, PropertySpec.prototype);

    /**
     * Numeric Property
     * @param name
     * @constructor
     */
    var NumericProperty = function(name){
        this.name = name;
        this.type = _INTEGER;
    }

    _.extend(NumericProperty.prototype, PropertySpec.prototype);

    /**
     * String Property
     * @param name
     * @constructor
     */
    var StringProperty = function(name){
        this.name = name;
        this.type = _STRING;
    }

    _.extend(StringProperty.prototype, PropertySpec.prototype)

    /**
     * Currency Property
     * @param name
     * @constructor
     */
    var CurrencyProperty = function(name){
        this.name  = name;
        this.type  = "Currency";
    }

    _.extend(CurrencyProperty.prototype, PropertySpec.prototype)


    var IntegerProperty = function(name){
        this.name  = name;
        this.type  = "Integer";
    }

    _.extend(IntegerProperty.prototype, PropertySpec.prototype)


    var NumericProperty = function(name){
        this.name  = name;
        this.type  = "Integer";
    }

    _.extend(NumericProperty.prototype, PropertySpec.prototype)


    var DateTimeProperty = function(name){
        this.name  = name;
        this.type  = _DATETIME;
    }

    _.extend(DateTimeProperty.prototype, PropertySpec.prototype)

    var GeoProperty = function(name){
        this.name  = name;
        this.type  = _GEO;
    }

    _.extend(GeoProperty.prototype, PropertySpec.prototype)


    var LinkProperty = function(name){
        this.name  = name;
        this.type  = _LINK;
    }

    _.extend(LinkProperty.prototype, PropertySpec.prototype)

    /*
     * Type Mixins
     */
    function bindProperty(p){
        var name = p.name;
        delete p.name;
        this.properties[name] = p;
        return this;
    }

    /**
     * Index
     */
    var IndexSpec  = function(keys){
        this.keys       = keys || [];
    }

    IndexSpec.prototype = {

        pushKey: function(key){
            this.keys.push(key);
        },

        asUnique: function(){
            this.options        = this.options || {}
            this.options.unique = true;
        }

    }

    /**
     * Property Binding
     */
    var PropertyBinding = {

        bindProperty: function(p){
            bindProperty.apply(this, arguments)
            return this;
        },

        withObject: function(name){
            return this.bindProperty(new PropertySpec(name))
        },

        withBoolean: function(name){
            return this.bindProperty(new BooleanProperty(name))
        },

        withNumber: function(name){
            return this.bindProperty(new NumericProperty(name))
        },

        withInteger: function(name){
            return this.bindProperty(new IntegerProperty(name))
        },

        withDateTime: function(name){
            return this.bindProperty(new DateTimeProperty(name))
        },

        withString: function(name){
            return this.bindProperty(new StringProperty(name))
        },

        withStringArray: function(name){
            return this.bindProperty(new StringProperty(name).asList())
        },

        withCurrency: function(name){
            return this.bindProperty(new CurrencyProperty(name))
        },

        withGeo: function(name){
            return this.bindProperty(new GeoProperty(name))
        },

        withLink: function(name){
            return this.bindProperty(new LinkProperty(name))
        },

        withType: function(name, type){
            return  this.bindProperty(new PropertySpec(name).ofType(type).asList())
        },

        withTypeArray: function(name, type){
            return this.bindProperty(new PropertySpec(name).ofType(type).asList())
        },

        mixinProperties: function(set){
            _.extend(this.properties, set.properties)
            return this;
        }

    }

    var PropertySet = function(){
        this.properties = []
    }

    _.extend(PropertySet.prototype, PropertyBinding, {
        /* Property Sets */
    })


    /**
     * Embedded Type
     * @param name
     * @constructor
     */
    var EmbeddedTypeSpec  = function(name, ns){
        this.name       = name;
        if(ns) this.inNamespace(ns)
        this.embedded   = true;
        this.properties = {};
    }

    _.extend(EmbeddedTypeSpec.prototype, PropertyBinding, {

        inNamespace: function(ns){
            this.namespace = ns;
            return this;
        },

        merge: function(spec){
            _.extend(this, spec);
        },

        toJSON: function(delimiter){
            return toJSON.apply(this, arguments)
        },

        toString: function(){
            return "embedded:" + this.name;
        }
    });

    /**
     * ResourceSpec
     *
     * @param name
     * @constructor
     *
     */
    var ResourceSpec = function(name, ns){
        this.name       = name;
        if(ns) this.inNamespace(ns)
        this.storedAs(name)
        this.properties = {};
    }

    _.extend(ResourceSpec.prototype, PropertyBinding, {

        merge: function(spec){
            _.extend(this, spec);
            return this;
        },

        storedAs: function(name){
            this.storage_name = name;
            return this;
        },

        inNamespace: function(ns){
            this.namespace = ns;
            return this;
        },

        documentedAt: function(url){
            this.documentation = url;
            return this;
        },

        asImmutable: function(flag){
            this.immutable = true;
            return this;
        },

        asExtensible: function(flag){
            this.extensible = true;
            return this;
        },

        asEncapsulated: function(flag){
            this.encapsulated = true;
            return this;
        },

        toJSON: function(delimiter){
            return toJSON.apply(this, arguments)
        },

        toString: function(){
            return "resource:" + this.name;
        }
    })

    //A Domain is a set of conceptually related objects
    var DomainSpec = function(namespace){
        this.namespace      = namespace;
        this.resources      = {}
    }

    DomainSpec.prototype = {

        createResource: function(name){
            this.ensureResources();

            if(_.contains(this.resources, name)) {
                return this[resource];
            }

            this.resources[name] = new ResourceSpec(name, namespace);

            return this.resources[name];
        },

        hasResource: function(name){
            this.ensureResources();
            return _.contains(this.resources, name)
        },

        getResource: function(name){
            this.ensureResources();
            return this.resources[name];
        },

        pushResource: function(resource){
            this.ensureResources();
            resource.inNamespace(this.namespace);
            this.resources[resource.name] = resource;
            return this;
        },

        ensureResources: function(){
            this.resources = this.resources || {};
        },

        toJSON: function(delimiter){
            return toJSON.apply(this, arguments)
        },

        toString: function(){
            return "domain:" + this.namespace;
        }

    }

    function createProperty(name, type){
        return new PropertySpec(name).ofType(type);
    }

    return {

        //Containers
        Domain          : DomainSpec
        , Resource        : ResourceSpec
        , Embedded        : EmbeddedTypeSpec
        , Index           : IndexSpec
        , PropertySet     : PropertySet

        //Scalar Properties
        //, createProperties: createProperty

        , Property        : PropertySpec
        , BooleanProperty : BooleanProperty
        , IntegerProperty : IntegerProperty
        , NumericProperty : NumericProperty
        , StringProperty  : StringProperty
        , CurrencyProperty: CurrencyProperty
        , DateTimeProperty: DateTimeProperty
        , GeoProperty     : DateTimeProperty
        , LinkProperty    : LinkProperty

        , namespaces   : {
            SYSTEM   : "core"
          , PLATFORM : "base"
        }

    }
})()