/**
 *
 * Set of standard rest callbacks
 *
 * @type {{noop: Function, status: Function, headers: Function, entity: Function, full: Function, selectCallback: Function}}
 */

;module.exports = function(options){
    'use strict'

    options = options || {};

    var util    = require('util');
    var http    = require('http');
    var chalk   = require('chalk');
    var _       = require('../utils');
    var converter = require('json-2-csv');

    var inspect = require('eyes').inspector({
        styles: {                 // Styles applied to stdout
            all:     'cyan',      // Overall style applied to everything
            label:   'underline', // Inspection labels, like 'array' in `array: [1, 2, 3]`
            other:   'inverted',  // Objects which don't have a literal representation, such as functions
            key:     'bold',      // The keys in object literals, like 'a' in `{a: 1}`
            special: 'red',      // null, undefined...
            string:  'yellow',
            number:  'green',
            bool:    'green',      // true false
            regexp:  'green'      // /\d+/
        },
        pretty: true,             // Indent object literals
        hideFunctions: false,     // Don't output functions at all
        stream: process.stdout,   // Stream to write to, or null
        maxLength: 4048           // Truncate output if longer
    });

    function lookupStatus(code){
        return (code.toString()) ? http.STATUS_CODES[code.toString()] : '';
    }

    function getColorfulStatus(response){
        var code = response.statusCode;
        if(code < 200){
            return chalk.black.bgWhite(' Ok ' + code + ' ' + lookupStatus(code) + ' ');
        } else if(code >= 200 && code < 300){
            return chalk.black.bgGreen(' Success ' + code + ' ' + lookupStatus(code) + ' ');
        } else if(code >= 300 && code < 400){
            return chalk.black.bgBlue(' Failed ' + code + ' ' + lookupStatus(code) + ' ');
        } else if(code >= 400 && code < 500){
            return chalk.black.bgRed(' Failed ' + code + ' ' + lookupStatus(code) + ' ');
        } else {
            return chalk.black.bgYellow('Problem ' + code + ' ');
        }
    }

    function reportError(err){
        console.log(chalk.white.bgRed(' Connection Failed '));
        console.log(err);
    }

    function present(content, overloads){
        overloads = overloads ||{};

        if(options.drop){
            delete content[options.drop];
        }

        if(options.status){
            if(options.color){
                console.log(getColorfulStatus(overloads.status));
            } else {
                console.log(overloads.status);
            }
        }
        if(options.color){
            inspect(content);
        } else {
            console.log(content);
        }
    }

    return {
        /**
         * Ignore the object
         * @param err
         * @param response
         * @param body
         */
        noop: function noopView(err, response, body) {},

        ping: function pingView(err, response, body) {
            if(err){
                return reportError(err);
            } else {
                if(response.statusCode === 200){
                    console.log(chalk.black.bgGreen(' Available '));
                } else {
                    console.log(chalk.black.bgYellow('Responding with status ' + response.statusCode));
                }
            }
        },

        auth: function authView(err, response, body) {
            if(err){
                reportError(err);
            } else {
                if(response.statusCode == 200){
                    console.log(chalk.black.bgGreen(' Authenticated '));
                } else if(response.statusCode == 400){
                    console.log(chalk.black.bgRed(' Error '));
                } else if(response.statusCode == 401){
                    console.log(chalk.black.bgRed(' Invalid Credentials '));
                } else if(response.statusCode == 403){
                    console.log(chalk.black.bgRed(' Forbidden '));
                } else if(response.statusCode == 404){
                    console.log(chalk.black.bgYellow(' Resource Unavailable '));
                } else {
                    console.log(chalk.black.bgYellow(response.statusCode));
                }
            }
        },

        status : function statusView(err, response, body) {
            if(err){
                reportError(err);
            } else {
                console.log(getColorfulStatus(response));
            }
        },

        headers: function headersView(err, response, body) {
            if(err){
                reportError(err);
            } else {
                present({
                    status  : response.statusCode,
                    headers : response.headers
                },{status: response});
            }
        },

        location: function locationView(err, response, body) {
            if(err){
                reportError(err);
            } else {
                var location = 'location header unavailable';
                try{
                    if(response.headers.location) {
                        location = response.headers.location;
                    }
                } catch (err){}
                console.log(location);
            }
        },

        csv: function csvView(err, response, body) {
            if(err){
                reportError(err);
            } else {
                if(response.statusCode !== 200){
                    console.log('unexpected return type ' + response.statusCode);
                } else {
                    if(!_.isString(body)){
                        console.log('unable to parse the result set');
                    } else {
                        body = JSON.parse(body);
                        converter.json2csv(body, function (err, csv) {
                            if (err) {
                                console.log(err);
                            }
                            console.log(csv);
                        });
                    }
                }
            }
        },

        id: function idView(err, response, body) {
            var location = 'location header unavailable';
            if(err){
                reportError(err);
            } else {
                try{
                    if(response.headers.location) {
                        location = response.headers.location;
                        location = location.substr(location.lastIndexOf('/') + 1)
                    }
                } catch (err){}
                return console.log(location);
            }
        },

        entity : function entityView(err, response, body) {
            if(err){
                return reportError(err);
            }
            try{
                body = JSON.parse(body);
                present(body, {status: response.status});
            } catch (err) {
                console.log('Invalid Response. Could not parse as JSON');
                present({
                    status  : response.statusCode,
                    headers : response.headers
                },{status: response});
                console.log(body);
            }
        },

        full : function fullView(err, response, body) {
            if(err){
                reportError(err);
            } else {
                present({
                    status  : response.statusCode,
                    headers : response.headers,
                    entity  : body
                }, {status: response});
            }
        },

        /**
         * The whole thing
         * @param err
         * @param response
         * @param body
         */
        all : function allView(err, response, body) {
            if(err){
                reportError(err);
            } else {
                present(response, {status: response})
            }
        }
    }
}